#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "vulkan_setup.h"

#include "vkload.h"

#include <stdio.h>		// TODO remove

static int compare_device_types(
		VkPhysicalDeviceType lhs,
		VkPhysicalDeviceType rhs)
{
	/* ascending priority */
	static VkPhysicalDeviceType const order[] =
	{
		VK_PHYSICAL_DEVICE_TYPE_OTHER,
		VK_PHYSICAL_DEVICE_TYPE_CPU,
		VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU,
		VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU,
		VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU
	};

	int lhs_index = -1;

	for(unsigned i = 0; i < sizeof order / sizeof *order; ++i)
		if(lhs == order[i])
		{
			lhs_index = i;
			break;
		}

	int rhs_index = -1;

	for(unsigned i = 0; i < sizeof order / sizeof *order; ++i)
		if(rhs == order[i])
		{
			rhs_index = i;
			break;
		}

	return rhs_index - lhs_index;
}

static int compare_present_modes(
		VkPresentModeKHR lhs,
		VkPresentModeKHR rhs)
{
	printf("%u %u\n", lhs, rhs);
	/* ascending priority */
	static VkPresentModeKHR const order[] =
	{
		VK_PRESENT_MODE_IMMEDIATE_KHR,
		VK_PRESENT_MODE_FIFO_RELAXED_KHR,
		VK_PRESENT_MODE_FIFO_KHR,
		VK_PRESENT_MODE_MAILBOX_KHR,
		VK_PRESENT_MODE_SHARED_CONTINUOUS_REFRESH_KHR,
		VK_PRESENT_MODE_SHARED_DEMAND_REFRESH_KHR,
	};

	int lhs_index = -1;

	for(unsigned i = 0; i < sizeof order / sizeof *order; ++i)
		if(lhs == order[i])
		{
			lhs_index = i;
			break;
		}

	int rhs_index = -1;

	for(unsigned i = 0; i < sizeof order / sizeof *order; ++i)
		if(rhs == order[i])
		{
			rhs_index = i;
			break;
		}

	return rhs_index - lhs_index;
}

bool vulkan_setup(struct vkload *g)
{
	/* horrible spaghetti code, kind of by choice */

	VkPhysicalDevice physical_device;

	uint32_t graphics_queue_family_index;
	uint32_t present_queue_family_index;

	g->instance = VK_NULL_HANDLE;
	g->surface = VK_NULL_HANDLE;
	g->device = VK_NULL_HANDLE;
	g->swapchain = VK_NULL_HANDLE;

	/* instance setup */
	{
		VkApplicationInfo const app_info =
		{
			.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO,
			.pApplicationName = PACKAGE,
			.applicationVersion = VK_MAKE_VERSION(0, 0, 0),
			.pEngineName = "No Engine",
			.engineVersion = VK_MAKE_VERSION(0, 0, 0),
			.apiVersion = VK_MAKE_VERSION(1, 0, 0)
		};

		char const *const extensions[] =
		{
			VK_KHR_XLIB_SURFACE_EXTENSION_NAME,
			"VK_KHR_surface",
			VK_EXT_DEBUG_UTILS_EXTENSION_NAME
		};

		char const *const enabledLayers[] =
		{
#ifndef NDEBUG
			/* run debug builds with validation */
			"VK_LAYER_KHRONOS_validation"
#endif
		};

		VkInstanceCreateInfo const info =
		{
			.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
			.pApplicationInfo = &app_info,
			.enabledExtensionCount = sizeof extensions / sizeof *extensions,
			.ppEnabledExtensionNames = extensions,
			.enabledLayerCount = sizeof enabledLayers / sizeof *enabledLayers,
			.ppEnabledLayerNames = enabledLayers
		};

		VkResult rc = vkCreateInstance(&info, NULL, &g->instance);

		if(rc != VK_SUCCESS)
			return false;
	}

	/* surface setup */
	{
		VkXlibSurfaceCreateInfoKHR const info =
		{
			.sType = VK_STRUCTURE_TYPE_XLIB_SURFACE_CREATE_INFO_KHR,
			// .pNext
			// .flags
			.dpy = g->display,
			.window = g->window
		};

		VkResult rc = vkCreateXlibSurfaceKHR(g->instance, &info, NULL, &g->surface);

		if(rc != VK_SUCCESS)
			return false;
	}

	/* pick physical device */
	{
		VkResult rc;

		/* get list of physical devices */
		uint32_t device_count = 0;

		rc = vkEnumeratePhysicalDevices(g->instance, &device_count, NULL);
		if(rc != VK_SUCCESS)
			return false;

		if(device_count == 0)
			return false;

		VkPhysicalDevice devices[device_count];

		uint32_t const saved_device_count = device_count;

		rc = vkEnumeratePhysicalDevices(g->instance, &device_count, devices);
		if(rc != VK_SUCCESS)
			return false;

		VkPhysicalDevice best_device = VK_NULL_HANDLE;
		VkPhysicalDeviceType best_type;

		for(uint32_t i = 0; i < saved_device_count; ++i)
		{
			VkPhysicalDevice const device = devices[i];

			/* checklist */
			bool have_graphics_queue_family_index = false;
			bool have_present_queue_family_index = false;
			bool have_surface_format = false;
			bool have_present_mode = false;

			/* selected values for current device */
			uint32_t device_graphics_queue_family_index;
			uint32_t device_present_queue_family_index;
			VkSurfaceFormatKHR device_surface_format;
			VkPresentModeKHR device_present_mode;
			VkSurfaceTransformFlagBitsKHR device_transform;
			uint32_t device_image_count;
			VkExtent2D device_extent;

			VkPhysicalDeviceProperties prop;

			vkGetPhysicalDeviceProperties(device, &prop);

			VkPhysicalDeviceType const device_type = prop.deviceType;

			/* for each device, get queue families supported */
			uint32_t family_count = 0;
			vkGetPhysicalDeviceQueueFamilyProperties(
					device,
					&family_count,
					NULL);

			if(family_count == 0)
				continue;

			VkQueueFamilyProperties family_properties[family_count];

			uint32_t const saved_family_count = family_count;

			vkGetPhysicalDeviceQueueFamilyProperties(
					device,
					&family_count,
					family_properties);

			for(uint32_t j = 0; j < saved_family_count; ++j)
			{
				VkBool32 present;

				/* check for graphics support */
				bool const graphics = 
					(family_properties[j].queueFlags & VK_QUEUE_GRAPHICS_BIT);

				/* check for present support */
				rc = vkGetPhysicalDeviceSurfaceSupportKHR(
						device,
						j,
						g->surface,
						&present);

				if(rc != VK_SUCCESS)
					continue;

				if(graphics)
				{
					device_graphics_queue_family_index = j;
					have_graphics_queue_family_index = true;
				}

				if(present)
				{
					device_present_queue_family_index = j;
					have_present_queue_family_index = true;
				}

				if(have_graphics_queue_family_index &&
						have_present_queue_family_index &&
						have_surface_format)
				{
					best_device = device,
					best_type = device_type;
					graphics_queue_family_index = device_graphics_queue_family_index;
					present_queue_family_index = device_present_queue_family_index;
					g->surface_format = device_surface_format;
					g->present_mode = device_present_mode;
					g->image_count = device_image_count;
					g->transform = device_transform;
					g->extent = device_extent;
					break;
				}
			}

			if(device != best_device)
				continue;

			/* skip if new device would be worse than the one we already have */
			if(best_device != VK_NULL_HANDLE)
				if(compare_device_types(best_type, device_type) < 0)
					continue;

			/* check surface capabilities? */
			VkSurfaceCapabilitiesKHR surface_capabilities;

			rc = vkGetPhysicalDeviceSurfaceCapabilitiesKHR(
					device,
					g->surface,
					&surface_capabilities);

			if(rc != VK_SUCCESS)
				continue;

			device_image_count = surface_capabilities.minImageCount + 1;
			if(surface_capabilities.maxImageCount != 0 &&
					surface_capabilities.maxImageCount < device_image_count)
				device_image_count = surface_capabilities.maxImageCount;

			device_transform = surface_capabilities.currentTransform;
			device_extent = surface_capabilities.currentExtent;

			uint32_t surface_formats_count = 0;
			rc = vkGetPhysicalDeviceSurfaceFormatsKHR(
					device,
					g->surface,
					&surface_formats_count,
					NULL);
			if(rc != VK_SUCCESS)
				continue;

			if(surface_formats_count == 0)
				continue;

			VkSurfaceFormatKHR surface_formats[surface_formats_count];

			uint32_t const saved_surface_formats_count = surface_formats_count;

			vkGetPhysicalDeviceSurfaceFormatsKHR(
					device,
					g->surface,
					&surface_formats_count,
					surface_formats);

			for(uint32_t j = 0; j < saved_surface_formats_count; ++j)
			{
				if(surface_formats[j].format == VK_FORMAT_B8G8R8A8_SRGB &&
						surface_formats[j].colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR)
				{
					/* good choice */
					device_surface_format = surface_formats[j];
					have_surface_format = true;
					break;
				}
			}

			if(!have_surface_format)
			{
				/* fallback */
				device_surface_format = surface_formats[0];
				have_surface_format = true;
				/* TODO prefer another device? */
			}

			uint32_t present_modes_count = 0;

			vkGetPhysicalDeviceSurfacePresentModesKHR(
					device,
					g->surface,
					&present_modes_count,
					NULL);

			if(present_modes_count == 0)
				continue;

			VkPresentModeKHR present_modes[present_modes_count];

			uint32_t const saved_present_modes_count = present_modes_count;

			vkGetPhysicalDeviceSurfacePresentModesKHR(
					device,
					g->surface,
					&present_modes_count,
					present_modes);

			for(uint32_t j = 0; j < saved_present_modes_count; ++j)
			{
				if(!have_present_mode ||
						compare_present_modes(
								device_present_mode,
								present_modes[j]) > 0)
				{
					device_present_mode = present_modes[j];
					have_present_mode = true;
				}
			}
		}

		if(best_device == VK_NULL_HANDLE)
			/* no device capable of rendering to surface */
			return false;

		physical_device = best_device;
	}

	/* logical device setup */
	{
		float const queue_priorities[] =
		{
			1.0f
		};

		VkDeviceQueueCreateInfo const queue_info[] =
		{
			/* this will be passed partially if the two family indexes are
			 * the same, so the count is not computed later and needs to
			 * be adjusted if anything changes here
			 */
			{
				.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
				.queueFamilyIndex = graphics_queue_family_index,
				.queueCount = sizeof queue_priorities / sizeof *queue_priorities,
				.pQueuePriorities = queue_priorities
			},
			{
				.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
				.queueFamilyIndex = present_queue_family_index,
				.queueCount = sizeof queue_priorities / sizeof *queue_priorities,
				.pQueuePriorities = queue_priorities
			},
		};

		/* "compute" the count of used queue_info members */
		uint32_t const queue_info_count =
			(graphics_queue_family_index == present_queue_family_index) ? 1 : 2;

		VkPhysicalDeviceFeatures const features =
		{
			0,
		};

		char const *const extensions[] =
		{
			VK_KHR_SWAPCHAIN_EXTENSION_NAME
		};

		char const *const enabledLayers[] =
		{
#ifndef NDEBUG
			/* run debug builds with validation */
			"VK_LAYER_KHRONOS_validation"
#endif
		};

		VkDeviceCreateInfo const info =
		{
			.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO,
			.queueCreateInfoCount = queue_info_count,
			.pQueueCreateInfos = queue_info,
			.pEnabledFeatures = &features,
			.enabledExtensionCount = sizeof extensions / sizeof *extensions,
			.ppEnabledExtensionNames = extensions,
			.enabledLayerCount = sizeof enabledLayers / sizeof *enabledLayers,
			.ppEnabledLayerNames = enabledLayers,
		};

		VkResult rc = vkCreateDevice(physical_device, &info, NULL, &g->device);

		if(rc != VK_SUCCESS)
			return false;
	}

	return true;

	/* get queues */
	{
		vkGetDeviceQueue(
				g->device,
				graphics_queue_family_index,
				0,
				&g->graphics_queue);
		vkGetDeviceQueue(
				g->device,
				present_queue_family_index,
				0,
				&g->present_queue);
	}

	/* set up swapchain */
	{
		VkSwapchainCreateInfoKHR const info =
		{
			.sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,
			.surface = g->surface,
			.minImageCount = g->image_count,
			.imageFormat = g->surface_format.format,
			.imageColorSpace = g->surface_format.colorSpace,
			.imageExtent = g->extent,
			.imageArrayLayers = 1,
			.imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
			.imageSharingMode = VK_SHARING_MODE_EXCLUSIVE,
			.queueFamilyIndexCount = 0,
			.pQueueFamilyIndices = NULL,
			.preTransform = g->transform,
			.compositeAlpha = VK_COMPOSITE_ALPHA_INHERIT_BIT_KHR,
			.presentMode = g->present_mode,
			.clipped = VK_TRUE,
			.oldSwapchain = VK_NULL_HANDLE
		};

		VkSwapchainKHR swapchain;

		VkResult rc = vkCreateSwapchainKHR(g->device, &info, NULL, &swapchain);

		if(rc != VK_SUCCESS)
			return false;

		g->swapchain = swapchain;
	}

	/* get swapchain images */
	{
		uint32_t images_count = 0;
		vkGetSwapchainImagesKHR(g->device, g->swapchain, &images_count, NULL);
		vkGetSwapchainImagesKHR(g->device, g->swapchain, &images_count, g->images);
	}

	return true;
}

void vulkan_teardown(struct vkload *g)
{
	if(g->swapchain != VK_NULL_HANDLE)
		vkDestroySwapchainKHR(g->device, g->swapchain, NULL);
	if(g->device != VK_NULL_HANDLE)
		vkDestroyDevice(g->device, NULL);
	if(g->surface != VK_NULL_HANDLE)
		vkDestroySurfaceKHR(g->instance, g->surface, NULL);
	if(g->instance != VK_NULL_HANDLE)
		vkDestroyInstance(g->instance, NULL);
}
