#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "x11_setup.h"

#include "vkload.h"

#include <X11/Xutil.h>

bool x11_setup(struct vkload *g)
{
	g->display = NULL;
	g->window = None;

	{
		Display *display = XOpenDisplay(NULL);

		if(!display)
			return false;

		g->display = display;
	}

	{
		Display *display = g->display;

		int const screen = DefaultScreen(display);
		unsigned long const black = BlackPixel(display, screen);
		unsigned long const white = WhitePixel(display, screen);

		Window window = XCreateSimpleWindow(
				display,
				DefaultRootWindow(display),
				0, 0,
				300, 300,
				0,
				black,
				white);

		/* failures are reported as unrecoverable X11 errors */
		g->window = window;
	}

	{
		XSizeHints *size_hints = XAllocSizeHints();

		// TODO fill out hints

		XSetStandardProperties(
				g->display,
				g->window,
				PACKAGE,
				PACKAGE,
				None,
				g->argv,
				g->argc,
				size_hints);

		XFree(size_hints);
	}

	XMapRaised(g->display, g->window);

	XSync(g->display, False);

	return true;
}

void x11_teardown(struct vkload *g)
{
	if(g->window != None)
		XDestroyWindow(g->display, g->window);

	if(g->display)
		XCloseDisplay(g->display);
}
